/*
 * MainViewer.cpp
 *
 */

#include "MainViewer.h"
#include "ViewerPool.h"
#include "Scene.h"
#include <QtGui>
#include <QGLViewer/manipulatedFrame.h>

using namespace qglviewer;

/*****************************************************************************
 MainViewer initialization
 *****************************************************************************/

MainViewer::MainViewer(Scene *s, ViewerPool *p, QWidget *v, int type, QWidget* parent, const QGLWidget* shareWidget) :
	QGLViewer(parent, shareWidget), scene(s), pool(p), sectionViewer(v) {
	connect(scene, SIGNAL(radiusChanged(float)), this, SLOT(setSceneRadius(float)));
	connect(scene, SIGNAL(radiusChanged(void)), this, SLOT(showEntireScene(void)));
	connect(scene, SIGNAL(sceneChanged(void)), this, SLOT(updateGL(void)));

	if (type < 3) {
		// Move camera according to viewer type (on X, Y or Z axis)
		camera()->setPosition(Vec((type == 0) ? 1.0 : 0.0, (type == 1) ? 1.0 : 0.0, (type == 2) ? 1.0 : 0.0));
		camera()->lookAt(sceneCenter());

		camera()->setType(Camera::ORTHOGRAPHIC);
		camera()->showEntireScene();

		// Forbid rotation
		WorldConstraint* constraint = new WorldConstraint();
		constraint->setRotationConstraintType(AxisPlaneConstraint::FORBIDDEN);
		camera()->frame()->setConstraint(constraint);
	}
	// The ManipulatedFrame will be used to position the clipping plane
	setManipulatedFrame(scene->manipulatedFrame());

	// Disable some predefined shortcuts
	setShortcut(STEREO, 0);
	setShortcut(EDIT_CAMERA, 0);

	// Set help descriptions
	setKeyDescriptions();
}

void MainViewer::init() {
	/*
	// Light setup
	glEnable(GL_LIGHT1);

	// Light default parameters
	const GLfloat light_ambient[4] = { 1.0, 1.0, 1.0, 1.0 };
	const GLfloat light_specular[4] = { 1.0, 1.0, 1.0, 1.0 };
	const GLfloat light_diffuse[4] = { 1.0, 1.0, 1.0, 1.0 };

	glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 3.0);
	glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 180.0);
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.5f);
	glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.3f);
	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.2f);
	glLightfv(GL_LIGHT1, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse);
	*/

	// Set radius scene and update camera
	setSceneRadius(scene->radius());
	camera()->showEntireScene();

	// Enable plane clipping
	glEnable(GL_CLIP_PLANE0);
}

/*****************************************************************************
 View rendering
 *****************************************************************************/

void MainViewer::draw() {
	/*
	// Place light at camera position
	const Vec cameraPos = camera()->position();
	const GLfloat pos[4] = { (GLfloat) cameraPos[0], (GLfloat) cameraPos[1], (GLfloat) cameraPos[2], 1.0f };
	glLightfv(GL_LIGHT1, GL_POSITION, pos);

	// Orient light along view direction
	glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, camera()->viewDirection());
	*/

	// Draw the scene
	scene->draw();
}

/*****************************************************************************
 Help descriptions
 *****************************************************************************/

QString MainViewer::helpString() const {
	QString text("<h2>3D Wood MainViewer</h2>");
	text += "Hold the <b>Control</b> key pressed down while using the mouse to modify the plane orientation ";
	text += "(left button) and position (right button) and to interactively see the clipped result.<br><br>";
	return text;
}

void MainViewer::setKeyDescriptions() {
	setKeyDescription(Qt::Key_C, tr("Loads a configuration file"));
	setKeyDescription(Qt::Key_O, tr("Loads a mesh (.obj) in the scene"));
	setKeyDescription(Qt::Key_Q, tr("Opens the wood section viewer"));
	setKeyDescription(Qt::Key_S, tr("Shows all viewers"));
	setKeyDescription(Qt::Key_M, tr("Masks the current viewer"));
}

/*****************************************************************************
 Keyboard events
 *****************************************************************************/

void MainViewer::keyPressEvent(QKeyEvent *event) {
	if (event->modifiers() != Qt::NoModifier) {
		QGLViewer::keyPressEvent(event);
		return;
	}
	switch (event->key()) {
	case Qt::Key_O: // Load a mesh
		loadMesh();
		break;
	case Qt::Key_C: // Load configuration file
		loadParameters();
		break;
	case Qt::Key_Q: // Open wood section viewer
		sectionViewer->show();
		sectionViewer->raise();
		break;
	case Qt::Key_S: // Show all viewers
		pool->showAll();
		break;
	case Qt::Key_M: // Mask current viewer
		pool->hide(this);
		break;
	default:
		QGLViewer::keyPressEvent(event);
	}
}

/*****************************************************************************
 Actions
 *****************************************************************************/

void MainViewer::loadMesh() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load a mesh (.obj)"), QDir::currentPath());
	scene->loadMesh(fileName);
}

void MainViewer::loadParameters() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load a configuration file"), QDir::currentPath());
	scene->loadParameters(fileName);
}
