/*
 * PlaneViewer.cpp
 *
 */

#include "PlaneViewer.h"
#include "WoodSection.h"
#include "Scene.h"
#include <QtGui>
#include <QGLViewer/frame.h>

using namespace qglviewer;

/*****************************************************************************
 Viewer initialization
 *****************************************************************************/

PlaneViewer::PlaneViewer(Scene* s, QWidget* m, QWidget* parent, const QGLWidget* shareWidget) :
	QGLViewer(parent, shareWidget), scene(s), mainViewer(m) {
	connect(scene, SIGNAL(radiusChanged(float)), this, SLOT(setSceneRadius(float)));
	connect(scene, SIGNAL(planeModified(void)), this, SLOT(updateCamera(void)));
	connect(scene, SIGNAL(sceneChanged(void)), this, SLOT(updateGL(void)));

	// Move camera according to viewer type (on X, Y or Z axis)
	camera()->setType(Camera::ORTHOGRAPHIC);

	// Change some predefined shortcuts
	setShortcut(STEREO, 0);
	setShortcut(ANIMATION, 0);
	setShortcut(EDIT_CAMERA, 0);
	setShortcut(INCREASE_FLYSPEED, 0);
	setShortcut(DECREASE_FLYSPEED, 0);
	setMouseBinding(Qt::LeftButton, NO_CLICK_ACTION);
	setMouseBinding(Qt::ShiftModifier + Qt::LeftButton, CAMERA, NO_MOUSE_ACTION);
	setMouseBinding(Qt::ControlModifier + Qt::LeftButton, CAMERA, NO_MOUSE_ACTION);

	// Set help descriptions
	setKeyDescriptions();
}

void PlaneViewer::init() {
	// Set radius scene and update camera
	setSceneRadius(scene->radius());
	updateCamera();

	// Enable plane clipping
	glEnable(GL_CLIP_PLANE0);
}

/*****************************************************************************
 View rendering
 *****************************************************************************/

void PlaneViewer::updateCamera() {
	Vec origin = plane()->inverseCoordinatesOf(Vec(0, 0, -5 * scene->radius()));
	Vec direction = plane()->inverseTransformOf(Vec(0, 0, 1));
	Vec up = plane()->inverseTransformOf(Vec(0, 1, 0));
	camera()->setPosition(origin);
	camera()->setViewDirection(direction);
	camera()->setUpVector(up);
	camera()->fitSphere(plane()->position(), scene->radius());
}

void PlaneViewer::drawWithNames() {
	scene->drawWithNames();
}

void PlaneViewer::draw() {
	scene->draw();
}

/*****************************************************************************
 Help descriptions
 *****************************************************************************/

QString PlaneViewer::helpString() const {
	QString text("<h2>3D Wood Section</h2>");
	text += "Hold the <b>Control</b> key pressed down while using the mouse to modify the plane orientation ";
	text += "(left button) and position (right button) and to interactively see the clipped result.<br><br>";
	return text;
}

void PlaneViewer::setKeyDescriptions() {
	setMouseBindingDescription(Qt::LeftButton, tr("Selects the nearest vertex/Moves selected vertex"));
	setMouseBindingDescription(Qt::LeftButton + Qt::ControlModifier, tr("Adds a new active vertex"));
	setMouseBindingDescription(Qt::LeftButton + Qt::ShiftModifier, tr("Adds a new inactive vertex"));

	setKeyDescription(Qt::Key_Return, tr("Runs the simulation process"));
	setKeyDescription(Qt::Key_Q, tr("Opens the main viewers window"));
	setKeyDescription(Qt::Key_V, tr("Toggles polygon vertices display"));
	setKeyDescription(Qt::Key_B, tr("Toggles polygon border display"));
	setKeyDescription(Qt::Key_X, tr("Toggles skeleton edges display"));
	setKeyDescription(Qt::Key_W, tr("Toggles wireframe display of the texture"));
	setKeyDescription(Qt::Key_1, tr("Toggles active cells information display"));
	setKeyDescription(Qt::Key_2, tr("Toggles former active cells information display"));
	setKeyDescription(Qt::Key_S, tr("Saves the current set of polygons"));
	setKeyDescription(Qt::Key_L, tr("Loads a set of polygons"));
	setKeyDescription(Qt::Key_C, tr("Loads a configuration file"));
	setKeyDescription(Qt::Key_N, tr("Creates a new active group"));
	setKeyDescription(Qt::Key_I, tr("Creates a new inactive group"));
	setKeyDescription(Qt::Key_R, tr("Removes selected vertex"));
	setKeyDescription(Qt::Key_U, tr("Unselects selected vertex"));
	setKeyDescription(Qt::Key_Plus, tr("Increase current generation number"));
	setKeyDescription(Qt::Key_Minus, tr("Decrease current generation number"));
	setKeyDescription(Qt::Key_Delete, tr("Clears the current set of polygons"));
}

/*****************************************************************************
 Keyboard events
 *****************************************************************************/

void PlaneViewer::keyPressEvent(QKeyEvent *event) {
	if (event->key() == Qt::Key_Question && event->modifiers().testFlag(Qt::ShiftModifier))
		QGLViewer::keyPressEvent(event);
	else if (event->modifiers().testFlag(Qt::ControlModifier))
		QGLViewer::keyPressEvent(event);
	else if (event->modifiers().testFlag(Qt::AltModifier))
		QGLViewer::keyPressEvent(event);
	else
		switch (event->key()) {
		case Qt::Key_V: // Toggle vertices info
			scene->woodSection()->toggleInfo(WoodSection::Vertices);
			break;
		case Qt::Key_B: // Toggle border info
			scene->woodSection()->toggleInfo(WoodSection::Border);
			break;
		case Qt::Key_X: // Toggle skeleton info
			scene->woodSection()->toggleInfo(WoodSection::Skeleton);
			break;
		case Qt::Key_W: // Toggle wireframe info
			scene->woodSection()->toggleInfo(WoodSection::Wireframe);
			break;
		case Qt::Key_1: // Toggle active cells info
			scene->woodSection()->toggleInfo(WoodSection::Active);
			break;
		case Qt::Key_2: // Toggle former active cells info
			scene->woodSection()->toggleInfo(WoodSection::Former);
			break;
		case Qt::Key_S: // Save current polygons
			saveSection();
			break;
		case Qt::Key_L: // Load polygon file
			loadSection();
			break;
		case Qt::Key_C: // Load configuration file
			loadParameters();
			break;
		case Qt::Key_R: // Remove selected vertex
			scene->woodSection()->removeSelection();
			setSelectedName(-1);
			break;
		case Qt::Key_U: // Unselect current vertex
			scene->woodSection()->select(-1);
			setSelectedName(-1);
			break;
		case Qt::Key_Minus: // Decrease generation number
			scene->diffIteration(-1);
			break;
		case Qt::Key_Plus: // Increase generation number
			scene->diffIteration(1);
			break;
		case Qt::Key_N: // Create new active polygon
			scene->woodSection()->createGroup(true);
			break;
		case Qt::Key_I: // Create new inactive polygon
			scene->woodSection()->createGroup(false);
			break;
		case Qt::Key_Delete: // Clear image
			scene->woodSection()->clear();
			break;
		case Qt::Key_Q: // Open main viewer
			mainViewer->show();
			mainViewer->raise();
			break;
		case Qt::Key_Return: // Run the simulation
			scene->woodSection()->growth(scene->parameters());
			scene->computeRadius();
			break;
		default:
			QGLViewer::keyPressEvent(event);
		}
}

#include <iostream>
using namespace std;

void PlaneViewer::mousePressEvent(QMouseEvent *event) {
	if (event->button() == Qt::LeftButton)
		switch (event->modifiers()) {
		case Qt::NoModifier: // Select nearest vertex
			scene->woodSection()->selectNearest(toPlane(event->pos()));
			break;
		case Qt::ControlModifier: // Add new active vertex
			scene->woodSection()->createVertex(toPlane(event->pos()));
			break;
		case Qt::ShiftModifier: // Add new inactive vertex
			scene->woodSection()->createVertex(toPlane(event->pos()), false);
			break;
		default:
			QGLViewer::mousePressEvent(event);
		}
	else
		QGLViewer::mousePressEvent(event);
}

void PlaneViewer::mouseMoveEvent(QMouseEvent *event) {
	if (event->buttons().testFlag(Qt::LeftButton) && event->modifiers() == Qt::NoModifier)
		scene->woodSection()->moveSelection(toPlane(event->pos()));
	else
		QGLViewer::mouseMoveEvent(event);
}

/*****************************************************************************
 Actions
 *****************************************************************************/

void PlaneViewer::loadSection() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load a wood section outline"), QDir::currentPath());
	scene->loadSection(fileName);
}

void PlaneViewer::saveSection() {
	QString fileName = QFileDialog::getSaveFileName(this, tr("Save current wood section outline"), QDir::currentPath());
	scene->saveSection(fileName);
}

void PlaneViewer::loadParameters() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load a configuration file"), QDir::currentPath());
	scene->loadParameters(fileName);
}

void PlaneViewer::postSelection(const QPoint& point) {
	Q_UNUSED(point);
	scene->woodSection()->select(selectedName());
}

/*****************************************************************************
 Plane projections
 *****************************************************************************/

const Frame* PlaneViewer::plane() const {
	return scene->woodSection()->plane();
}

Vec PlaneViewer::convertClickToPlane(const QPoint &pixel) const {
	/* Normal code would use something like :
	 *
	 * camera()->convertClickToLine(pixel, orig, dir);
	 * Vec normal = plane->inverseTransformOf(Vec(0, 0, 1));
	 * double param = ((plane->position() - orig) * normal) / (dir * normal);
	 *
	 * But since here normal == dir, we can simplify a bit
	 */

	Vec orig, dir;
	camera()->convertClickToLine(pixel, orig, dir);
	double param = ((plane()->position() - orig) * dir) / dir.squaredNorm();
	return orig + param * dir;
}

QPointF PlaneViewer::projectOnPlane(const Vec &pos) const {
	Vec coord = plane()->coordinatesOf(pos);
	return QPointF(coord.x, coord.y);
}

QPointF PlaneViewer::toPlane(const QPoint &pixel) const {
	return projectOnPlane(convertClickToPlane(pixel));
}
