/*
 * Cell.cpp
 *
 */

#include "Cell.h"
#include "Geometry.h"
#include "Random.h"

/*****************************************************************************
 Miscellaneous
 *****************************************************************************/

void Cell::smoothSpeed() {
	double alpha = angle(left->position - this->position);
	double beta = angle(right->position - this->position);
	speed = fromPolar(length(speed), mod((alpha - beta) / 2.0, Pi) + beta);
}

void Cell::checkSpeed() {
	if (not (det(left->position - position, speed) > 0 && det(speed, right->position - position) > 0))
		smoothSpeed();
}

/*****************************************************************************
 Cell creation routines
 *****************************************************************************/

Cell * Cell::makeNewCell(Cell *parent, double angularStdev, double lengthStdev, double distanceLimit) {
	Cell *newCell;
	if (parent->state == Cell::Alive) {
		double newAngle = rng.normal(angle(parent->speed), angularStdev);
		double newLength = qMin(rng.normal(length(parent->speed), lengthStdev), distanceLimit);
		newLength = qMax(newLength, distanceLimit / 2.);
		newCell = new Cell(parent->position + parent->speed, fromPolar(newLength, newAngle), parent->radius + 1,
			parent->clockWise);
	} else
		newCell = new Cell(*parent);
	newCell->parent = QVector<Cell *> (1, parent);
	parent->child = newCell;
	return newCell;
}

Cell * Cell::makeBetweenArc(Cell *left, Cell *right) {
	QPointF position = interpolatePositionArc(*left, *right);
	QPointF speed = interpolateSpeedArc(*left, *right);
	double radius = qRound(0.5 * (left->radius + right->radius));
	State state = (left->state == Dead && right->state == Dead ? Dead : Alive);
	Cell *middle = new Cell(position, speed, radius, left->clockWise, state, left, right);
	bindThree(left, middle, right);
	return middle;
}

Cell * Cell::makeBetweenFlat(Cell *left, Cell *right) {
	QPointF position = interpolatePositionFlat(*left, *right);
	QPointF speed = interpolateSpeedFlat(*left, *right);
	int radius = (left->radius + right->radius) / 2;
	State state = (left->state == Dead && right->state == Dead ? Dead : Alive);
	Cell *middle = new Cell(position, speed, radius, left->clockWise, state, left, right);
	bindThree(left, middle, right);
	return middle;
}

Cell * Cell::mergeCells(Cell *left, Cell *right) {
	QPointF position = interpolatePositionFlat(*left, *right);
	QPointF speed = interpolateSpeedArc(*left, *right);
	int radius = (left->radius + right->radius) / 2;
	State state = (left->state == Dead && right->state == Dead ? Dead : Alive);
	Cell *middle = new Cell(position, speed, radius, left->clockWise, state, left, right);
	bindThree(left->left, middle, right->right);
	middle->parent << left->parent << right->parent;
	delete left;
	delete right;
	for (int i = 0; i < middle->parent.size(); ++i)
		middle->parent[i]->child = middle;
	return middle;
}

/*****************************************************************************
 Cell insertion between two other cells
 *****************************************************************************/

Cell * Cell::insertLeftOf(Cell *right, const QPointF &position) {
	// TODO: Factor code with makeBetweenFlat
	Cell *left = right->left;
	QPointF speed = interpolateSpeedFlat(*left, *right);
	int radius = (left->radius + right->radius) / 2;
	State state = (left->state == Dead && right->state == Dead ? Dead : Alive);
	Cell *middle = new Cell(position, speed, radius, left->clockWise, state, left, right);
	bindThree(left, middle, right);
	return middle;
}

/*****************************************************************************
 Speed and position interpolation
 *****************************************************************************/

QPointF Cell::interpolateSpeedArc(const Cell &left, const Cell &right) {
	double meanLength = 0.5 * (length(left.speed) + length(right.speed));
	return meanLength * normalized(left.speed + right.speed);
}

// The new speed is deduced from the position of the adjacent cells only,
// And its length is averaged from the adjacent cells as well
QPointF Cell::interpolateSpeedFlat(const Cell &left, const Cell &right) {
	QPointF meanPos = 0.5 * (left.position + right.position);
	double alpha = angle(left.position - meanPos);
	double beta = angle(right.position - meanPos);
	double meanLength = 0.5 * (length(left.speed) + length(right.speed));
	return fromPolar(meanLength, mod((alpha - beta) / 2.0, Pi) + beta);
}

// A more complex interpolation that place the new cell on a circle deduce from each cells' speed
QPointF Cell::interpolatePositionArc(const Cell &left, const Cell &right) {
	QPointF meanPos = 0.5 * (left.position + right.position);
	QPointF segment = right.position - left.position;

	// Trajectory of the 2 cells, and their perpendicular bisector
	QLineF leftTraj(left.position, left.position + left.speed);
	QLineF rightTraj(right.position, right.position + right.speed);
	QLineF perpTraj(meanPos, meanPos + QPointF(segment.y(), -segment.x()));

	// Intersection with the perpendicular bisector of "left" and "right"
	QPointF leftInter, rightInter;
	leftTraj.intersect(perpTraj, &leftInter);
	rightTraj.intersect(perpTraj, &rightInter);

	// If one intersection is beyond the segment, return meanPos
	//if (intersectSegm(left.position, right.position, leftInter, rightInter))
	if (det(leftInter - left.position, segment) > 0 || det(rightInter - right.position, segment) > 0)
		return meanPos;
	else {
		// Translate the middle cell to a circle passing through "left" and "right"
		// At the intersection with their perpendicular bisector
		QPointF meanInter = 0.5 * (leftInter + rightInter);
		double radius = length(meanInter - left.position);
		Q_ASSERT(qAbs(length(meanInter - left.position) - length(meanInter - right.position)) < fEpsilon);
		intersectCircle(perpTraj, meanInter, radius, leftInter, rightInter);

		// Return the point in the circle that is closest to "meanPos"
		if (lengthSquared(leftInter - meanPos) < lengthSquared(rightInter - meanPos))
			return leftInter;
		else
			return rightInter;
	}
}

QPointF Cell::interpolatePositionFlat(const Cell &left, const Cell &right) {
	return 0.5 * (left.position + right.position);
}

