/*
 * Cell.h
 *
 */

#ifndef CELL_H_
#define CELL_H_

#include <QPointF>
#include <QVector>

// Struct definition
class Cell
{
	// Internal state
	enum State
	{
		Alive, Dead
	};

public:
	Cell(const QPointF &pos, const QPointF &speed = QPointF(), int radius = 0, bool clockWise = true, State state =
		Alive, Cell *left = 0, Cell *right = 0);

	QPointF position;
	QPointF speed;
	int radius;
	int index;
	int group;
	bool clockWise;

	Cell *left;
	Cell *right;
	Cell *child;
	QVector<Cell *> parent;

public:
	// Miscellaneous
	bool isDead() const; //! Returns true iff the cell is considered dead
	void kill(); //! Set the cell's state to 'Dead'
	void detach() const; //! Removes the current cell to its parents' neighborhood
	void checkSpeed(); //! Ensure the cell's speed is consistent
	void smoothSpeed(); //! Reinterpolate the cell's speed from the neighboring vertices

	// Static helper function
	static void bindTwo(Cell *left, Cell *right);
	static void bindThree(Cell *left, Cell *middle, Cell *right);
	static void checkSpeed(Cell *cell1, Cell *cell2);

	// Public named constructors
	static Cell * makeNewCell(Cell *parent, double angularStdev, double lengthStdev, double distanceLimit);
	static Cell * makeBetweenArc(Cell *left, Cell *right);
	static Cell * makeBetweenFlat(Cell *left, Cell *right);
	static Cell * mergeCells(Cell *left, Cell *right);
	static Cell * insertLeftOf(Cell *cell, const QPointF &pos);

private:
	// Private static interpolation functions
	static QPointF interpolateSpeedArc(const Cell &left, const Cell &right);
	static QPointF interpolateSpeedFlat(const Cell &left, const Cell &right);
	static QPointF interpolatePositionArc(const Cell &left, const Cell &right);
	static QPointF interpolatePositionFlat(const Cell &left, const Cell &right);

	State state;
};

/*****************************************************************************
 Cell inline methods
 *****************************************************************************/

inline Cell::Cell(const QPointF &position, const QPointF &speed, int radius, bool clockWise, State state,
	Cell *left, Cell *right) :
	position(position), speed(speed), radius(radius), index(-1), group(-1), clockWise(clockWise), left(left), right(
		right), child(0), state(state) {
}

inline bool Cell::isDead() const {
	return state == Dead;
}

inline void Cell::kill() {
	state = Dead;
}

inline void Cell::detach() const {
	for (int i = 0; i < parent.size(); ++i)
		parent[i]->child = 0;
}

/*****************************************************************************
 Cell inline helper functions
 *****************************************************************************/

inline void Cell::bindTwo(Cell *left, Cell *right) {
	left->right = right;
	right->left = left;
}

inline void Cell::bindThree(Cell *left, Cell *middle, Cell *right) {
	bindTwo(left, middle);
	bindTwo(middle, right);
}

inline void Cell::checkSpeed(Cell *cell1, Cell *cell2) {
	cell1->checkSpeed();
	cell2->checkSpeed();
}

#endif /* CELL_H_ */
