/*
 * CellGroup.cpp
 *
 */

#include "CellGroup.h"
#include "Geometry.h"
#include "Cell.h"

/*****************************************************************************
 Constructor, copy constructor and destructor
 *****************************************************************************/

CellGroup::CellGroup() :
	allDead(false), doNotKeep(false), orientation(false) {
}

CellGroup::CellGroup(QVector<Cell *> cells, bool orientation) :
	start(cells.first()), size(cells.size()), allDead(false), doNotKeep(false), orientation(orientation) {
	for (int i = 0; i < size; ++i) {
		cells[i]->right = cells[(i + 1) % size];
		cells[(i + 1) % size]->left = cells[i];
	}
}

CellGroup::CellGroup(Cell *start, int size, bool orientation, const QRectF &box) :
	start(start), size(size), box(box), allDead(false), doNotKeep(false), orientation(orientation) {
}

/*****************************************************************************
 Miscellaneous methods
 *****************************************************************************/

int CellGroup::computeSize() const {
	int counter = 0;
	for (Cell *cell = start->right;; cell = cell->right) {
		++counter;
		if (cell == start)
			break;
	}
	return counter;
}

bool CellGroup::computeOrientation() const {
	double signedArea = 0;
	for (Cell *cell = start->right;; cell = cell->right) {
		signedArea += det(cell->position, cell->right->position);
		if (cell == start)
			break;
	}
	return signedArea > 0;
}

void CellGroup::updateCellsOrientation() const {
	for (Cell *cell = start->right;; cell = cell->right) {
		cell->clockWise = orientation;
		if (cell == start)
			break;
	}
}

/*****************************************************************************
 Update the bounding box
 *****************************************************************************/

void CellGroup::updateBoundingBox() {
	box = QRectF(start->position, start->position);
	for (Cell *cell = start->right;; cell = cell->right) {
		box |= QRectF(box.topLeft(), cell->position);
		if (cell == start)
			break;
	}
}

/*****************************************************************************
 Set cells' group index
 *****************************************************************************/

void CellGroup::setGroupIndex(int index) {
	for (Cell *cell = start->right;; cell = cell->right) {
		cell->group = index;
		if (cell == start)
			break;
	}
}

/*****************************************************************************
 Check for group consistency
 *****************************************************************************/

// Check if the polygon has the expected orientation
// And check if all its cells are dead or not
void CellGroup::checkConsistency() {
	allDead = true;
	double signedArea = 0;
	for (Cell *cell = start->right;; cell = cell->right) {
		allDead &= cell->isDead();
		signedArea += det(cell->position, cell->right->position);
		if (cell == start)
			break;
	}
	if (orientation != (signedArea > 0))
		doNotKeep = true;
}

/*****************************************************************************
 Erase cells in the group
 *****************************************************************************/

int CellGroup::eraseBetween(Cell *left, Cell *right) {
	Cell *tmp = left->right;
	if (left == right)
		return 0;
	int eraseCounter = 0;
	while (tmp != right) {
		++eraseCounter;
		if (start == tmp)
			start = right;
		tmp = tmp->right;
		tmp->left->detach();
		delete (tmp->left);
	}
	Cell::bindTwo(left, right);
	size -= eraseCounter;
	return eraseCounter;
}

int CellGroup::eraseAliveBetween(Cell *left, Cell *right) {
	Cell *tmp = left->right;
	if (left == right)
		return 0;
	int eraseCounter = 0;
	while (tmp != right) {
		Cell *next = tmp->right;
		if (not tmp->isDead()) {
			Cell::bindTwo(tmp->left, tmp->right);
			++eraseCounter;
			if (start == tmp)
				start = right;
			tmp->detach();
			delete tmp;
		}
		tmp = next;
	}
	size -= eraseCounter;
	return eraseCounter;
}
