/*
 * Mesh.cpp
 *
 */

#include "Mesh.h"
#include "Color.h"
#include <QtOpenGL>

Mesh::Mesh() :
	hasNormals(false), hasTexCoords(false) {
}

Mesh::~Mesh() {
}

/*****************************************************************************
 Radius estimation
 *****************************************************************************/

float Mesh::computeRadius() const {
	qreal radius = 0;
	for (int i = 0; i < vertices.size(); ++i)
		radius = qMax(radius, vertices[i].length());
	return radius;
}

/*****************************************************************************
 Drawing function
 *****************************************************************************/

void Mesh::draw() const {
	// Enable and load arrays
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, vertices.constData());

	if (hasNormals) {
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT, 0, normals.constData());
	}

	if (hasTexCoords) {
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(3, GL_FLOAT, 0, texCoords.constData());
	}

	// Draw faces
	GLfloat faceColor[4];
	qSetColor(faceColor, QColor(Qt::darkGreen));

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, faceColor);
	glDrawElements(GL_TRIANGLES, 3 * faces.size(), GL_UNSIGNED_INT, faces.constData());

	// Restore states
	glDisableClientState(GL_VERTEX_ARRAY);
	if (hasNormals)
		glDisableClientState(GL_NORMAL_ARRAY);
	if (hasTexCoords)
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

/*****************************************************************************
 Miscellaneous
 *****************************************************************************/

void drawCube() {
	GLfloat vertices[] = { 1, 1, 1, -1, 1, 1, -1, -1, 1, 1, -1, 1, // v0-v1-v2-v3
		1, 1, 1, 1, -1, 1, 1, -1, -1, 1, 1, -1, // v0-v3-v4-v5
		1, 1, 1, 1, 1, -1, -1, 1, -1, -1, 1, 1, // v0-v5-v6-v1
		-1, 1, 1, -1, 1, -1, -1, -1, -1, -1, -1, 1, // v1-v6-v7-v2
		-1, -1, -1, 1, -1, -1, 1, -1, 1, -1, -1, 1, // v7-v4-v3-v2
		1, -1, -1, -1, -1, -1, -1, 1, -1, 1, 1, -1 }; // v4-v7-v6-v5
	GLfloat normals[] = { 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, // v0-v1-v2-v3
		1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, // v0-v3-v4-v5
		0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, // v0-v5-v6-v1
		-1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, // v1-v6-v7-v2
		0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, // v7-v4-v3-v2
		0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1 }; // v4-v7-v6-v5
	GLfloat colors[] = { 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 1, // v0-v1-v2-v3
		1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, // v0-v3-v4-v5
		1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, // v0-v5-v6-v1
		1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, // v1-v6-v7-v2
		0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, // v7-v4-v3-v2
		0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1 }; // v4-v7-v6-v5
	GLubyte indices[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23 };

	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);

	glNormalPointer(GL_FLOAT, 0, normals);
	glColorPointer(3, GL_FLOAT, 0, colors);
	glVertexPointer(3, GL_FLOAT, 0, vertices);

	glDrawElements(GL_QUADS, 24, GL_UNSIGNED_BYTE, indices);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
}
