#include "MeshImporter.h"
#include <QFile>

#include <iostream>
using namespace std;

bool MeshImporter::import(const QString &fileName, Mesh &mesh) {
	// Clear any past data
	clear();

	// Open file
	QFile file(fileName);
	if (not file.open(QIODevice::ReadOnly | QIODevice::Text))
		return false;

	QTextStream inputStream(&file);

	mHasNormals = false;
	mHasTexCoords = false;

	// Read vertex info and faces
	if (not readData(inputStream, mesh))
		return false;

	// Close file
	file.close();

	// Set properties
	mesh.hasNormals = mHasNormals;
	mesh.hasTexCoords = mHasTexCoords;

	return true;
}

void MeshImporter::clear() {
	mHasTexCoords = false;
	mHasNormals = false;
	mPositions.clear();
	mNormals.clear();
	mTexCoords.clear();
	mVertexMap.clear();
}

bool MeshImporter::readData(QTextStream &inputFile, Mesh &mesh) {
	while (not inputFile.atEnd()) {
		// Read the first and second characters
		QString element;
		inputFile >> element;

		if (element == "v") {
			mPositions.append(readVector3D(inputFile));
			skipLine(inputFile);
		} else if (element == "vn") {
			mHasNormals = true;
			mNormals.append(readVector3D(inputFile));
			skipLine(inputFile);
		} else if (element == "vt") {
			mHasTexCoords = true;
			QVector2D texCoord2 = readVector2D(inputFile);

			// Read an extra texture coordinate parameter if present
			qreal w = inputFile.readLine().toDouble();
			mTexCoords.append(QVector3D(texCoord2.x(), texCoord2.y(), w));
		} else if (element == "f") {
			// Initialize vertex map
			mVertexMap.resize(mPositions.size());
			readFace(inputFile, mesh);
		} else
			skipLine(inputFile);

		if (inputFile.status() != QTextStream::Ok)
			return false;
	}

	return true;
}

void MeshImporter::readFace(QTextStream &inputFile, Mesh &mesh) {
	Face face;
	QString line = inputFile.readLine().replace("//", "/0/");
	QTextStream inputLine(&line);

	// Read first face
	for (uint i = 0; i < 3; ++i) {
		IndexTriplet triplet;
		readTriplet(inputLine, triplet);
		face.vertices[i] = createVertex(triplet, mesh);
	}

	// Insert face
	mesh.faces.append(face);

	// Read next faces
	while (not inputLine.atEnd()) {
		face.vertices[1] = face.vertices[2];
		IndexTriplet triplet;
		readTriplet(inputLine, triplet);
		face.vertices[2] = createVertex(triplet, mesh);
		mesh.faces.append(face);
	}
}

void MeshImporter::skipLine(QTextStream &inputFile) {
	(void) inputFile.readLine();
}

QVector2D MeshImporter::readVector2D(QTextStream &inputFile) {
	qreal x, y;
	inputFile >> x >> y;

	return QVector2D(x, y);
}

QVector3D MeshImporter::readVector3D(QTextStream &inputFile) {
	qreal x, y, z;
	inputFile >> x >> y >> z;

	return QVector3D(x, y, z);
}

void MeshImporter::readTriplet(QTextStream &inputLine, IndexTriplet &triplet) {
	triplet.i2 = 0;
	triplet.i3 = 0;

	char separator;
	inputLine >> triplet.i1 >> separator;
	if (separator == '/') {
		inputLine >> triplet.i2 >> separator;
		if (separator == '/')
			inputLine >> triplet.i3;
	}
	inputLine.skipWhiteSpace();
}

uint MeshImporter::createVertex(const IndexTriplet &triplet, Mesh &mesh) {
	uint vertexIndex = triplet.i1 - 1;
	IndexPair pair(triplet.i2 - 1, triplet.i3 - 1);
	VertexMap::ConstIterator it = mVertexMap[vertexIndex].find(pair);

	if (it == mVertexMap[vertexIndex].end()) {
		mesh.vertices.append(mPositions[vertexIndex]);
		mesh.texCoords.append(triplet.i2 ? mTexCoords[pair.first] : QVector3D());
		mesh.normals.append(triplet.i3 ? mNormals[pair.second] : QVector3D(1, 0, 0));

		mVertexMap[vertexIndex][pair] = mesh.vertices.size() - 1;
		return mesh.vertices.size() - 1;
	}

	return it.value();
}
