/*
 * Scene.cpp
 *
 */

#include "Scene.h"
#include "Object.h"
#include "WoodSection.h"
#include "Parameters.h"
#include "MeshImporter.h"
#include "Mesh.h"

#include <QGLViewer/qglviewer.h>
#include <QGLViewer/manipulatedFrame.h>

using namespace qglviewer;

/*****************************************************************************
 Initialization
 *****************************************************************************/

Scene::Scene() :
	sceneRadius(10.f), params(new Parameters()), frame(new ManipulatedFrame()) {
	currentSection = new WoodSection(frame);
	params->showSeed();

	// Connect signals
	connect(frame, SIGNAL(modified(void)), this, SIGNAL(planeModified(void)));
	connect(frame, SIGNAL(modified(void)), this, SIGNAL(sceneChanged(void)));
	connect(currentSection, SIGNAL(objectChanged(void)), this, SIGNAL(sceneChanged(void)));

	/* Temporary default loading
	*  MeshImporter importer;
	*  Mesh *mesh = new Mesh;
	*  const QString fileName = "/home/tln/workspace/infographie/projet/viewer/objects/tree.obj";
	*  if (importer.import(fileName, *mesh)) {
	*  	objects.append(mesh);
	*	connect(mesh, SIGNAL(objectChanged(void)), this, SIGNAL(sceneChanged(void)));
	*
	* 	computeRadius();
	* 	sceneChanged();
	* }
	*/
}

Scene::~Scene() {
	for (int i = 0; i != objects.size(); ++i) {
		disconnect(objects[i], SIGNAL(objectChanged(void)), this, SIGNAL(sceneChanged(void)));
		delete objects[i];
	}
	delete frame;
	delete currentSection;
}

void Scene::computeRadius() {
	// Radius of the current wood section
	sceneRadius = currentSection->plane()->position().norm() + currentSection->computeRadius();

	// Other objects in the scene
	for (int i = 0; i != objects.size(); ++i)
		sceneRadius = qMax(sceneRadius, objects[i]->computeRadius());

	// Add one for safety
	sceneRadius += 1.f;
	radiusChanged(sceneRadius);
	radiusChanged();
}

/*****************************************************************************
 Mesh loading
 *****************************************************************************/

void Scene::loadMesh(const QString &fileName) {
	MeshImporter importer;
	Mesh *mesh = new Mesh;
	if (importer.import(fileName, *mesh)) {
		// Upon success, remove the other objects and add the new one
		for (int i = 0; i != objects.size(); ++i) {
			disconnect(objects[i], SIGNAL(objectChanged(void)), this, SIGNAL(sceneChanged(void)));
			delete objects[i];
		}

		objects.clear();
		objects.append(mesh);
		connect(mesh, SIGNAL(objectChanged(void)), this, SIGNAL(sceneChanged(void)));

		computeRadius();
		sceneChanged();
		return;
	} else
		delete mesh;
}

void Scene::loadSection(const QString &fileName) {
	currentSection->load(fileName);
	computeRadius();
	sceneChanged();
}

void Scene::saveSection(const QString &fileName) {
	currentSection->save(fileName);
}

void Scene::loadParameters(const QString &fileName) {
	params->load(fileName);
	params->showSeed();
	sceneChanged();
}

void Scene::diffIteration(int i) {
	params->growthIter = qMax(0, params->growthIter + i);
}

/*****************************************************************************
 Content rendering
 *****************************************************************************/

void Scene::drawWithNames() const {
	glPushMatrix();
	glMultMatrixd(currentSection->plane()->matrix());
	currentSection->drawWithNames(sceneRadius / 100.f);
	glPopMatrix();
}

void Scene::draw() const {
	glPushMatrix();
	glMultMatrixd(currentSection->plane()->matrix());

	// Since the Clipping Plane equation is multiplied by the current modelView, we can define a
	// constant equation (plane normal along Z and passing by the origin) since we are here in the
	// manipulatedFrame coordinates system (we glMultMatrixd() with the manipulatedFrame matrix()).
	static const GLdouble equation[] = { 0.0, 0.0, 1.0, 0.0 };
	glClipPlane(GL_CLIP_PLANE0, equation);

	// Draw a plane representation: Its normal...
	glColor3f(0.8f, 0.8f, 0.8f);

	// ...and a quad (with a slightly shifted z so that it is not clipped).
	float s = sceneRadius;
	glNormal3f(0, 0, 1.f);
	glBegin(GL_QUADS);
	glVertex3f(-s, -s, 0.1f);
	glVertex3f(-s, s, 0.1f);
	glVertex3f(s, s, 0.1f);
	glVertex3f(s, -s, 0.1f);
	glEnd();

	// Draw the wood section
	currentSection->draw(sceneRadius / 100.f);

	glPopMatrix();

	// Draw the other objects of the scene
	for (int i = 0; i != objects.size(); ++i)
		objects[i]->draw();
}
