/*
 * Geometry.h
 *
 */

#ifndef GEOMETRY_H_
#define GEOMETRY_H_

#include "StdDefinition.h"
#include "Triplet.h"

#include <cstdio>
#include <qmath.h>
#include <algorithm>

#include <QLinkedList>
#include <QPolygonF>
#include <QVector>
#include <QPointF>
#include <QLineF>
#include <QList>
#include <QPair>

/*****************************************************************************
 QPointF auxiliary functions
 *****************************************************************************/

inline qreal lengthSquared(const QPointF & p) {
	return p.x() * p.x() + p.y() * p.y();
}

inline qreal length(const QPointF & p) {
	return qSqrt(lengthSquared(p));
}

inline qreal dot(const QPointF & u, const QPointF & v) {
	return u.x() * v.x() + u.y() * v.y();
}

inline qreal det(const QPointF & u, const QPointF & v) {
	return u.x() * v.y() - u.y() * v.x();
}

inline qreal angle(const QPointF & p) {
	return qAtan2(-p.y(), p.x());
}

bool inline isReflex(const QPointF & a, const QPointF & b, const QPointF & c) {
	return det(b - a, c - a) > 0;
}

inline QPointF fromPolar(qreal length, qreal angle) {
	return QPointF(length * qCos(angle), -length * qSin(angle));
}

inline QPointF normalVector(const QPointF &p) {
	return QPointF(p.y(), -p.x());
}

inline QPointF normalized(const QPointF &p) {
	return p / length(p);
}

inline bool fuzzyCompare(const QPointF &p1, const QPointF &p2) {
	return qFuzzyCompare(p1.x(), p2.x()) && qFuzzyCompare(p1.y(), p2.y());
}

/*****************************************************************************
 Miscellaneous geometry tools
 *****************************************************************************/

// Return true iff p is in the triangle {a,b,c}
inline bool isInsideTriangle(const QPointF &a, const QPointF &b, const QPointF &c, const QPointF &p) {
	bool s1 = det(p - a, b - a) > 0;
	bool s2 = det(p - b, c - b) > 0;
	bool s3 = det(p - c, a - c) > 0;
	return s1 == s2 && s2 == s3;
}

// Return true iff p is in the triangle {a,b,c,d}
inline bool isInsideQuad(const QPointF &a, const QPointF &b, const QPointF &c, const QPointF &d, const QPointF &p) {
	bool s1 = det(p - a, b - a) > 0;
	bool s2 = det(p - b, c - b) > 0;
	bool s3 = det(p - c, a - c) > 0;
	bool s4 = det(p - a, d - a) > 0;
	bool s5 = det(p - d, c - d) > 0;
	return (s1 == s2 && s2 == s3) || (s3 == s4 && s4 == s5);
}

// Find the circumcenter of the triangle {a,b,c}
inline QLineF::IntersectType circumcenter(const QPointF &a, const QPointF &b, const QPointF &c, QPointF &center) {
	QLineF ab(a, b);
	QLineF bc(b, c);
	QLineF bisectAB = ab.normalVector().translated(0.5 * ab.dx(), 0.5 * ab.dy());
	QLineF bisectBC = bc.normalVector().translated(0.5 * bc.dx(), 0.5 * bc.dy());
	return bisectAB.intersect(bisectBC, &center);
}

// Intersection between a line and a circle
inline bool intersectCircle(const QLineF &line, const QPointF &center, double radius, QPointF &intersect1,
	QPointF &intersect2) {
	QPointF u = line.p2() - line.p1();
	QPointF v = line.p1() - center;
	double a = lengthSquared(u);
	double b = 2 * dot(u, v);
	double c = lengthSquared(v) - radius * radius;

	double delta = b * b - 4 * a * c;
	if (delta >= 0) {
		intersect1 = line.pointAt((-b + sqrt(delta)) / (2.0 * a));
		intersect2 = line.pointAt((-b - sqrt(delta)) / (2.0 * a));
		return true;
	}

	return false;
}

// Test whether the segments ]a,b[ && ]b,c[ intersect each others
inline bool intersectSegm(const QPointF &a, const QPointF &b, const QPointF &c, const QPointF &d) {
	double z = det(b - a, d - c);
	double x = det(c - a, b - a);
	double y = det(d - c, a - c);
	int sx = (x > 0 ? 1 : -1);
	int sy = (y > 0 ? 1 : -1);
	// Use only integers if possible
	return (z != 0 && x * z >= 0 && sx * x < sx * z && y * z >= 0 && sy * y < sy * z);
}

// Intersection between two lines
inline void intersectRay(const QPointF &a, const QPointF &u, const QPointF &b, const QPointF &v, double &t1, double &t2) {
	double x = det(a - b, v);
	double y = det(u, b - a);
	double z = -det(u, v);
	t1 = x / z;
	t2 = y / z;
}

// Return true iff the two quad overlap
inline bool overlapQuad(const QPointF &a1, const QPointF &a2, const QPointF &a3, const QPointF &a4, const QPointF &b1,
	const QPointF &b2, const QPointF &b3, const QPointF &b4) {
	return intersectSegm(a1, a2, b1, b2) || intersectSegm(a1, a2, b2, b3) || intersectSegm(a1, a2, b3, b4)
		|| intersectSegm(a1, a2, b4, b1) || intersectSegm(a2, a3, b1, b2) || intersectSegm(a2, a3, b2, b3)
		|| intersectSegm(a2, a3, b3, b4) || intersectSegm(a2, a3, b4, b1) || intersectSegm(a3, a4, b1, b2)
		|| intersectSegm(a3, a4, b2, b3) || intersectSegm(a3, a4, b3, b4) || intersectSegm(a3, a4, b4, b1)
		|| intersectSegm(a4, a1, b1, b2) || intersectSegm(a4, a1, b2, b3) || intersectSegm(a4, a1, b3, b4)
		|| intersectSegm(a4, a1, b4, b1);
}

/*****************************************************************************
 Triangulation of a polygon
 *****************************************************************************/

template<typename PolygonF>
void triangulatePolygon(const PolygonF &poly, QVector<Triplet> &answer) {
	int n = poly.size();
	int loopCount = 0;
	int unassigned = n;
	QVector<QPair<int, int> > hull;
	QVector<bool> mark(n, false);
	QLinkedList<int> notch;
	QLinkedList<int>::Iterator it;

	// Empty polygon
	if (poly.size() < 3)
		return;

	// Detect polygon orientation
	double signedArea = 0;
	for (int j = 0; j < n; ++j)
		signedArea += det(poly[j], poly[mod(j + 1, n)]);
	bool isClockWise = signedArea < 0;

	// Initialization
	for (int j = 0; j < n; ++j) {
		hull.append(qMakePair(mod(j - 1, n), mod(j + 1, n)));
		if (isReflex(poly[mod(j - 1, n)], poly[j], poly[mod(j + 1, n)]) == isClockWise)
			notch.append(j);
	}

	while (not notch.isEmpty() && unassigned > 2) {
		int i1 = notch.first();
		int i2 = hull[i1].second;
		int i3 = hull[i2].second;

		// Check if there is a notch inside the new polygon
		bool badTriangle = false;
		for (it = notch.begin() + 1; it != notch.end(); ++it) {
			if (*it == i2 || (*it != i3 && isInsideTriangle(poly[i1], poly[i2], poly[i3], poly[*it]))) {
				badTriangle = true;
				break;
			}
		}

		// Avoid infinite loops in case of ill-formed polygon
		if (++loopCount > notch.size())
			return;

		// Skip current vertex
		if (badTriangle) {
			notch.append(notch.first());
			notch.removeFirst();
			continue;
		}
		loopCount = 0;

		// Create the triangle
		answer.append(Triplet(i1, i2, i3));

		// Collapse the link
		--unassigned;
		mark[i2] = true;
		hull[i1].second = i3;
		hull[i3].first = i1;
		if (isReflex(poly[hull[i1].first], poly[i1], poly[i3]) != isClockWise)
			notch.removeOne(i1);
		if (isReflex(poly[i1], poly[i3], poly[hull[i3].second]) != isClockWise)
			notch.removeOne(i3);
	}

	// Triangulate the remaining convex polygon
	for (int i = 0; i < n; ++i) {
		if (not mark[i]) {
			for (int j = hull[i].second; hull[j].second != i; j = hull[j].second)
				answer.append(Triplet(i, j, hull[j].second));
			break;
		}
	}

	Q_ASSERT(answer.size() == poly.size() - 2);
}

/*****************************************************************************
 Convex Hull of a set of points
 *****************************************************************************/

// Counterclockwise order
struct CCWOrder
{
	CCWOrder(const QPointF & leftmost) :
		leftmost(leftmost) {
	}

	inline bool operator ()(const QPointF & a, const QPointF & b) const {
		qreal d = det(a - leftmost, b - leftmost);
		if (a == leftmost)
			return true;
		if (b == leftmost)
			return false;
		return (d < -fEpsilon || (d < fEpsilon && lengthSquared(a - leftmost) < lengthSquared(b - leftmost)));
	}

private:
	const QPointF leftmost;
};

// Graham scan, complexity O(n⋅log(n))
template<typename PolygonF>
void convexHull(PolygonF & poly, PolygonF & hull) {
	int i, j;

	// Search for leftmost vertex
	for (i = 1, j = 0; i < poly.size(); ++i)
		if (poly[i].x() < poly[j].x())
			j = i;
	std::sort(poly.begin(), poly.end(), CCWOrder(poly[j]));

	// Scan each point in counterclockwise order
	hull.clear();
	for (i = 0; i < poly.size(); ++i) {
		hull.push_back(poly[i]);
		// Pop vertices that are inside the hull
		while (hull.size() > 3 && isReflex(hull.end()[-3], hull.end()[-2], hull.end()[-1])) {
			hull.end()[-2] = hull.back();
			hull.pop_back();
		}
	}
	return;
}

/*****************************************************************************
 Partition into convex polygons
 *****************************************************************************/

// Class for the vertices incident to the diagonals
struct Vertex
{
	Vertex(int num = -1, int to = -1, Vertex *left = 0, Vertex *right = 0) :
		num(num), to(to), left(left), right(right) {
		if (left != 0)
			left->right = this;
		if (right != 0)
			right->left = this;
	}

	int num, to;
	Vertex *left;
	Vertex *right;

	inline int numLeft(int n) const {
		return (left == 0) ? mod(num - 1, n) : left->to;
	}

	inline int numRight(int n) const {
		return (right == 0) ? mod(num + 1, n) : right->to;
	}

	inline void collapse() const {
		if (left != 0)
			left->right = right;
		if (right != 0)
			right->left = left;
	}
};

// Class for the vertices on the remaining polygon
struct Element
{
	Element(int left = -1, int right = -1) :
		left(left), right(right), leftArc(0), rightArc(0) {
	}
	int left, right;
	Vertex *leftArc;
	Vertex *rightArc;
};

// Complexity O(n×r)
template<typename PolygonF>
void convexPartition(const PolygonF &poly, QList<QVector<int> > &answer) {
	int i = 0, n = poly.size();
	int loopCount = 0;
	QVector<Element> hull;
	QVector<QPair<Vertex*, Vertex*> > graph;
	QLinkedList<int> notch;
	QLinkedList<int>::Iterator it;

	// Empty polygon
	if (poly.size() < 3)
		return;

	// Initialization
	for (int j = 0; j < n; ++j) {
		hull.append(Element(mod(j - 1, n), mod(j + 1, n)));
		if (isReflex(poly[mod(j - 1, n)], poly[j], poly[mod(j + 1, n)]))
			notch.append(j);
	}

	while (not notch.isEmpty()) {
		i = notch.first();

		const QPointF & r = poly[hull[i].left];
		const QPointF & s = poly[i];
		const QPointF & t = poly[hull[i].right];

		// Add vertices in clockwise order
		int nRight = 2;
		int j = hull[i].right;
		for (; j != i; j = hull[j].right, ++nRight) {
			const QPointF & a = poly[hull[j].left];
			const QPointF & b = poly[j];
			const QPointF & c = poly[hull[j].right];
			if (isReflex(a, b, c) || isReflex(b, c, s) || isReflex(c, s, t))
				break;

			// Check if there is a notch inside the new polygon
			bool notchInside = false;
			for (it = notch.begin() + 1; it != notch.end(); ++it)
				if ((notchInside = isInsideTriangle(s, b, c, poly[*it])))
					break;
			if (notchInside)
				break;
		}

		// Add vertices in counter-clockwise order
		int nLeft = 2;
		int l = hull[i].left;
		for (; l != i; l = hull[l].left, ++nLeft) {
			const QPointF & a = poly[hull[l].left];
			const QPointF & b = poly[l];
			const QPointF & c = poly[hull[l].right];
			if (isReflex(a, b, c) || isReflex(s, a, b) || isReflex(r, s, a))
				break;

			// Check if there is a notch inside the new polygon
			bool notchInside = false;
			for (it = notch.begin() + 1; it != notch.end(); ++it)
				if ((notchInside = isInsideTriangle(a, b, s, poly[*it])))
					break;
			if (notchInside)
				break;
		}

		// Go to the side with minimum number of inserted vertices
		if (nLeft < nRight) {
			j = i;
			i = l;
			nRight = nLeft;
		}

		// Avoid infinite loops in case of ill-formed polygon
		if (++loopCount > notch.size())
			return;

		// Special case, the notch is to be skipped
		if (nRight <= 2) {
			notch.append(notch.first());
			notch.removeFirst();
			continue;
		} else
			loopCount = 0;

		// Add edge (i,j) to the graph
		graph.append(qMakePair(new Vertex(i, j, hull[i].leftArc, hull[i].rightArc), new Vertex(j, i, hull[j].leftArc,
			hull[j].rightArc)));
		hull[i].rightArc = graph.last().first;
		hull[j].leftArc = graph.last().second;

		// Collapse the link
		hull[i].right = j;
		hull[j].left = i;
		if (not isReflex(poly[hull[i].left], poly[i], poly[j]))
			notch.removeOne(i);
		if (not isReflex(poly[i], poly[j], poly[hull[j].right]))
			notch.removeOne(j);
	}

	// Merge non-essential edges
	QVector<bool> crossed(graph.size(), false);
	for (int k = 0; k < graph.size(); ++k) {
		const Vertex &e = *graph[k].first;
		const Vertex &f = *graph[k].second;

		// If the edge is essential ...
		if (isReflex(poly[e.numLeft(n)], poly[e.num], poly[e.numRight(n)]))
			continue;
		if (isReflex(poly[f.numLeft(n)], poly[f.num], poly[f.numRight(n)]))
			continue;

		// Remove edge k from the graph
		crossed[k] = true;
		e.collapse();
		f.collapse();
	}

	// Create the actual partition
	QPair<int, int> leap(0, 1);
	for (int k = 0; k < graph.size(); ++k) {
		if (crossed[k])
			continue;

		const Vertex &e = *graph[k].first;
		const Vertex &f = *graph[k].second;

		answer.append(QVector<int> (1, f.num));
		for (int j = e.num; j != f.num;) {
			answer.last().append(j);
			j = (j == leap.first ? leap.second : mod(j + 1, n));
		}
		leap = qMakePair(e.num, f.num);
	}

	// Additional loop for the last polygon
	answer.append(QVector<int> (1, leap.first));
	for (int j = leap.second; j != leap.first; j = mod(j + 1, n))
		answer.last().append(j);

	// Free the allocated structures
	for (int k = 0; k < graph.size(); ++k) {
		delete graph[k].first;
		delete graph[k].second;
	}
}

#endif /* GEOMETRY_H_ */
