/*
 * WoodCells.cpp
 *
 */

#include "WoodCells.h"
#include "Parameters.h"
#include "Geometry.h"
#include "Cell.h"
#include "CellGroup.h"
#include "CellIntersections.h"

#include <cstdio>
#include <QSet>
#include <QVector>

/*****************************************************************************
 Constructor, copy constructor and destructor
 *****************************************************************************/

WoodCells::WoodCells(const Parameters &params, const QVector<QPolygonF> &activePolys,
	const QVector<QPolygonF> &inactivePolys) :
	zeta(params.zeta), distanceLimit(params.distanceLimit), distanceSquared(distanceLimit * distanceLimit),
		angularStdev(params.angularStdev), lengthStdev(params.lengthStdev), growthIter(params.growthIter),
		groupCounter(0), stepPerCycle(params.stepPerCycle), colorEarlyWood(params.cEarly), colorLateWood(params.cLate) {
	for (int i = 0; i < 4; ++i)
		speed[i] = params.initialSpeed[i];

	// Initialize active groups of cells
	if (activePolys.isEmpty())
		populatePoint(QPointF(0, 0));
	for (int i = 0; i < activePolys.size(); ++i) {
		populatePoly(activePolys[i]);
		populateInactive(activePolys[i]);
	}

	// Initialize inactive groups of cells
	for (int i = 0; i < inactivePolys.size(); ++i)
		populateInactive(inactivePolys[i]);

	// Check correctness of the result
	for (int i = 0; i < active.size(); ++i) {
		mergeGroup(active[i]);
		eraseIntersections(active[i]);
		active[i].updateBoundingBox();
	}

	// Mark the elements of the current groups to be deleted later
	recordConfiguration();
	initial = active;
}

WoodCells::~WoodCells() {
	QSet<Cell *> mark;

	// Delete visible elements and check for memory corruption
	for (int i = 0; i < elements.size(); ++i) {
		if (mark.contains(elements[i]))
			printf("Corruption: %p\n", elements[i]);
		else
			delete elements[i];
		mark.insert(elements[i]);
	}

	// Delete invisible elements and check for memory corruption
	for (int i = 0; i < invisible.size(); ++i) {
		if (mark.contains(invisible[i]))
			printf("Corruption: %p\n", invisible[i]);
		else
			delete invisible[i];
		mark.insert(invisible[i]);
	}

	fflush(0);
}

/*****************************************************************************
 Initialization
 *****************************************************************************/

void WoodCells::populatePoint(const QPointF &orig) {
	QVector<Cell *> children;

	// Start with a minimal group of 4 cells
	QPointF dx = QPointF(distanceLimit / 2., 0.);
	QPointF dy = QPointF(0., distanceLimit / 2.);
	children.append(new Cell(orig + dx, QPointF(qMin(speed[0], distanceLimit), 0)));
	children.append(new Cell(orig + dy, QPointF(0, qMin(speed[1], distanceLimit))));
	children.append(new Cell(orig - dx, QPointF(-qMin(speed[2], distanceLimit), 0)));
	children.append(new Cell(orig - dy, QPointF(0, -qMin(speed[3], distanceLimit))));

	// Append to the set of active cells
	active.append(CellGroup(children));
	active.last().orientation = active.last().computeOrientation();
	active.last().updateCellsOrientation();
}

void WoodCells::populatePoly(const QPolygonF &poly) {
	// Special case : empty polygon or single-point polygon
	if (poly.isEmpty())
		return;
	else if (poly.size() == 1) {
		populatePoint(poly.first());
		return;
	}

	// Insert points on every edge
	QVector<Cell *> children;
	for (int i = 0, n = poly.size(); i < n; ++i) {
		QLineF edge(poly[i], poly[(i + 1) % n]);
		int m = ceil(edge.length() / distanceLimit);
		for (int j = 0; j < m; ++j)
			children.append(new Cell(edge.pointAt((double) j / m), QPointF(distanceLimit / 2., 0)));
	}

	// Append to the set of active cells
	active.append(CellGroup(children));
	active.last().orientation = active.last().computeOrientation();
	active.last().updateCellsOrientation();

	// Smooth each cell's speed
	for (Cell *cell = children.first()->right;; cell = cell->right) {
		cell->smoothSpeed();
		if (cell == children.first())
			break;
	}
}

void WoodCells::populateInactive(const QPolygonF &poly) {
	if (poly.size() < 2)
		return;

	// Insert points on every vertex
	QVector<Cell *> children(poly.size());
	for (int i = 0, n = poly.size(); i < n; ++i) {
		children[i] = new Cell(poly[n - i - 1]);
		children[i]->kill();
		invisible.append(children[i]);
	}

	// Append to the set of inactive cells
	inactive.append(CellGroup(children));
	inactive.last().updateBoundingBox();
}

/*****************************************************************************
 Growth algorithm's main loop
 *****************************************************************************/

void WoodCells::growth() {
	for (int k = 0; k < growthIter; ++k) {
		QVector<CellGroup> oldGroup;
		qSwap(oldGroup, active);

		// Initial breeding
		for (int i = 0; i < oldGroup.size(); ++i)
			initNewGroup(oldGroup[i]);

		// Sanity checks
		checkConfiguration();

		// Inter-generation overlap detection
		for (int i = 0; i < oldGroup.size(); ++i)
			readjustGroup(oldGroup[i]);

		// Garbage collection
		recordConfiguration();
	}

	printf("Simulation ended with %d cells.\n", elements.size());
}

/*****************************************************************************
 Apply heuristics to detect abnormalities
 *****************************************************************************/

void WoodCells::checkConfiguration() {
	for (int i = 0; i < active.size(); ++i) {
		splitIntersections(active[i], i);
		mergeGroup(active[i]);
		eraseIntersections(active[i]);
		active[i].updateBoundingBox();
		active[i].checkConsistency();
		active[i].setGroupIndex(++groupCounter);
	}

#ifdef DEBUG
	for (int i = 0; i < active.size(); ++i)
	Q_ASSERT(active[i].size == active[i].computeSize());
#endif

	for (int i = 0; i < active.size(); ++i)
		for (int j = i + 1; j < active.size(); ++j)
			checkCollisions(active[i], active[j]);

	for (int i = 0; i < active.size(); ++i)
		for (int j = 0; j < inactive.size(); ++j)
			checkCollisions(active[i], inactive[j]);
}

/*****************************************************************************
 Mark the cells for garbage collection
 *****************************************************************************/

void WoodCells::recordConfiguration() {
	for (int i = 0; i < active.size(); ++i) {
		for (Cell *cell = active[i].start->right;; cell = cell->right) {
			elements.append(cell);
			if (cell == active[i].start)
				break;
		}
	}
}

/*****************************************************************************
 New group initialization
 *****************************************************************************/

void WoodCells::initNewGroup(const CellGroup &group) {
	if (group.size < 4 || group.doNotKeep) {
		final.append(group);
		return;
	} else if (group.allDead) {
		inactive.append(group);
		return;
	}

#ifdef VERBOSE
	printf("- Initialization\n");
#endif

	// Generate a new group of active cells
	QVector<Cell *> children;
	for (Cell *cell = group.start->right;; cell = cell->right) {
		children.append(Cell::makeNewCell(cell, angularStdev, lengthStdev, distanceLimit));
		if (cell == group.start)
			break;
	}
	CellGroup newGroup(children, group.orientation);
	children.clear();

	// Check for new cells to be inserted
	for (Cell *cell = newGroup.start->right;; cell = cell->right) {
		while (not (cell->isDead() && cell->left->isDead()) && lengthSquared(cell->position - cell->left->position)
			> distanceSquared) {
			cell = Cell::makeBetweenArc(cell->left, cell);
			++newGroup.size;
		}
		if (cell == newGroup.start)
			break;
	}

	active.append(newGroup);
}

/*****************************************************************************
 Splitting a group into several parts
 *****************************************************************************/

void WoodCells::splitIntersections(CellGroup group, int i) {
#ifdef VERBOSE
	printf("- Splitting a group of size: %d\n", group.size);
#endif
	if (group.orientation)
		return;

	// Detect "pits" in the hull
	for (Cell *cell = group.start->right; group.size >= 4; cell = cell->right) {
		bool moveStart = false;
		double distanceMax = 0;
		Cell *tmp = cell->right;
		for (int index = 0; 2 * index < group.size;) {
			double distanceCurrent = length(tmp->position - cell->position);
			moveStart |= (tmp == group.start);
			if (2 * distanceCurrent < distanceMax && distanceCurrent < 2 * distanceLimit) {
				if (distanceMax < 2 * distanceLimit || index < 4) {
					// 1st case: small hull, delete it
					index -= group.eraseBetween(cell, tmp);
					distanceMax = distanceCurrent;
					Cell::checkSpeed(cell, tmp);
				} else {
					// 2nd case: long hull, split it
					group.size -= index;
					CellGroup newGroup(cell->right, index, group.orientation);

					Cell::bindTwo(tmp->left, cell->right);
					newGroup.size += bridgeBetween(tmp->left, cell->right);
					Cell::checkSpeed(tmp->left, cell->right);

					Cell::bindTwo(cell, tmp);
					Cell::checkSpeed(cell, tmp);
					active.append(newGroup);

					if (moveStart)
						group.start = tmp;
					break;
				}
			} else if (distanceCurrent > distanceMax)
				distanceMax = distanceCurrent;
			tmp = tmp->right;
			++index;
		}
		if (cell == group.start)
			break;
	}

	active[i] = group;
}

// Insert cells in the segment [left, right]
int WoodCells::bridgeBetween(Cell *left, Cell *right) const {
	int addCounter = 0;
	for (Cell *cell = left; cell != right; cell = cell->right) {
		while (lengthSquared(cell->position - cell->right->position) > distanceSquared) {
			Cell::makeBetweenFlat(cell, cell->right);
			++addCounter;
		}
	}
	return addCounter;
}

/*****************************************************************************
 Merging cells in a group
 *****************************************************************************/

void WoodCells::mergeGroup(CellGroup &group) {
#ifdef VERBOSE
	printf("- Merging\n");
#endif

	// Merge cells that would cross each others
	Cell *cell = group.start;
	for (int index = 0; index < group.size && group.size >= 4; ++index) {
		// TODO : Is it necessary ?
		if (cell->isDead())
			cell->right->smoothSpeed();
		else if (cell->right->isDead())
			cell->smoothSpeed();
		Cell::checkSpeed(cell, cell->right);
		if (length(cell->right->position - cell->position) < fEpsilon * fEpsilon || (not (cell->isDead()
			&& cell->right->isDead()) && intersectSegm(cell->position, cell->position + cell->speed,
			cell->right->position, cell->right->position + cell->right->speed))) {
			// Create a new intermediary cell and restore the links
			cell = Cell::mergeCells(cell, cell->right)->left;
			--group.size;
			index = 0;
		} else
			cell = cell->right;
	}

	group.start = cell;
}

/*****************************************************************************
 Self-intersection detection
 *****************************************************************************/

// TODO : Changer le test « overlapQuad » si l'interpolation est de type « Arc »
// TODO : Ce n'est peut-être pas nécessaire (grâce au test d'orientation ...)
bool overlapPatch(Cell *first, Cell *second) {
	if (first == second->right || second == first->right)
		return intersectSegm(first->position, first->position + first->speed, second->position, second->position
			+ second->speed);
	else
		return overlapQuad(first->position + first->speed, first->position, first->right->position,
			first->right->position + first->right->speed, second->position + second->speed, second->position,
			second->right->position, second->right->position + second->right->speed);
}

void WoodCells::eraseIntersections(CellGroup &group) const {
#ifdef VERBOSE
	printf("- Intersection\n");
#endif

	// Detect self intersections of the hull
	for (Cell *cell = group.start->right; group.size >= 4; cell = cell->right) {
		Cell *tmp = cell->right;
		for (int index = 1; tmp != cell && group.size >= 4;) {
			if (overlapPatch(cell, tmp)) {
				if (2 * index <= group.size) {
					group.eraseBetween(cell, tmp->right);
					Cell::checkSpeed(cell, cell->right);
					tmp = cell->right;
					index = 1;
				} else {
					group.eraseBetween(tmp, cell->right);
					Cell::checkSpeed(tmp, tmp->right);
					cell = tmp->right;
					index = group.size - 1;
				}
				group.start = cell->left;
			} else {
				tmp = tmp->right;
				++index;
			}
		}
		if (cell == group.start)
			break;
	}
}

/*****************************************************************************
 Collision detection
 *****************************************************************************/

void WoodCells::checkCollisions(CellGroup &group1, CellGroup &group2) {
	if (not group1.box.intersects(group2.box))
		return;
	Cell *a = group1.start;
	Cell *b = group2.start;
	CellIntersections intersections;
	for (int x = 0; x < group1.size; ++x, a = a->right) {
		if (not (group2.box.contains(a->position) || group2.box.contains(a->right->position)))
			continue;
		intersections.minParameter = 1.0;
		for (int y = 0; y < group2.size; ++y, b = b->right)
			if (not (fuzzyCompare(a->position, b->position) || fuzzyCompare(a->position, b->right->position)
				|| fuzzyCompare(a->right->position, b->position)
				|| fuzzyCompare(a->right->position, b->right->position)))
				if (intersectSegm(a->position, a->right->position, b->position, b->right->position)) {
					double t = det(a->position - b->position, b->right->position - b->position) / det(
						b->right->position - b->position, a->right->position - a->position);
					if (det(a->right->position - a->position, b->right->position - b->position) > 0)
						intersections.push(a, b->right, true, t);
					else
						intersections.push(a->right, b, false, t);
				}
		intersections.minorGather();
	}
	intersections.finalGather();

#ifdef VERBOSE
	printf("- Collisions detected: %d\n", intersections.size());
#endif

	for (int k = 0; k < intersections.size(); ++k) {
		QPair<Cell *, Cell *> goRight = intersections.interRight[k];
		QPair<Cell *, Cell *> goLeft = intersections.interLeft[k];
		Cell *x = goRight.first->left;
		Cell *y = goLeft.first->right;
		Cell *a = goLeft.second->left;
		Cell *b = goRight.second->right;

		// TODO
		QPointF interLeft;
		QPointF interRight;
		QLineF(x->position, x->right->position).intersect(QLineF(b->position, b->left->position), &interLeft);
		QLineF(a->position, a->right->position).intersect(QLineF(y->position, y->left->position), &interRight);

		group1.eraseAliveBetween(x, y);
		group2.eraseAliveBetween(a, b);
		x->right->kill();
		y->left->kill();
		a->right->kill();
		b->left->kill();

		// TODO : Insert neighbors ?

		/*
		 x->right->position = QLineF(interLeft, x->right->position).pointAt(0.9);
		 b->left->position = QLineF(interLeft, b->left->position).pointAt(0.9);
		 y->left->position = QLineF(interRight, y->left->position).pointAt(0.9);
		 a->right->position = QLineF(interRight, a->right->position).pointAt(0.9);
		 */
	}
}

/*****************************************************************************
 Avoid overlaps between generations
 *****************************************************************************/

void WoodCells::readjustGroup(CellGroup &oldGroup) {
	Cell *oldFirst(0), *oldLast(0);
	return;

	// Change groupStart to a valid parcel "corner"
	while (not oldGroup.start->child)
		oldGroup.start = oldGroup.start->right;

	for (oldFirst = oldGroup.start; oldLast != oldGroup.start; oldFirst = oldLast) {
		// Look for rightmost vertex of the parcel
		oldLast = oldFirst->right;
		while (not oldLast->child)
			oldLast = oldLast->right;

		// If parcel is not valid
		if (oldFirst->child->group != oldLast->child->group)
			continue;

		// Check whether parcel self-intersects
		bool noIntersection = true;
		for (Cell *level1 = oldFirst->child; noIntersection && level1 != oldLast->child; level1 = level1->right) {
			for (Cell *level0 = oldFirst; noIntersection && level0 != oldLast; level0 = level0->right) {
				if (intersectSegm(level1->position, level1->right->position, level0->position, level0->right->position))
					noIntersection = false;
			}
		}

		// If no intersection found, do not shrink current parcel
		if (noIntersection)
			continue;

		// TODO: If oldFirst or oldLast form a reflex angle ...
		// TODO: Special when there is an intersection with (oldFirst | oldLast) and its child
		if (oldFirst->right == oldLast) {
			QPointF middle = (oldFirst->child->position + oldLast->child->position) / 2.;
			int counter = 0;
			for (Cell *cell = oldFirst->child->right; cell != oldLast->child; cell = cell->right) {
				++counter;
				cell->position = middle;
			}
			printf("Cells aligned: %d\n", counter);
		} else {
			int counter = 0;
			Cell *target = oldFirst->right;
			for (Cell *cell = oldFirst->child->right; cell != oldLast->child; cell = cell->right) {
				++counter;
				cell->position = target->position;
				if (target != oldLast->left)
					target = target->right;
			}
			// Insert cells if needed
			int inserted = 0;
			for (Cell *cell = oldLast->child; target != oldLast->left; target = target->right)
				cell = Cell::insertLeftOf(cell, target->position), ++inserted;
			printf("Cells out of place: %d, inserted: %d\n", counter, inserted);
		}

		// TODO: Smooth speeds of newly displaced cells
	}
}
