/*
 * WoodCells.h
 *
 */

#ifndef WOODCELLS_H_
#define WOODCELLS_H_

#include "StdDefinition.h"
#include "Cell.h"
#include <QPair>
#include <QColor>
#include <QVector>
#include <QPolygonF>

class CellGroup;
class Parameters;
class WoodSkeleton;

/*****************************************************************************
 Class definition
 *****************************************************************************/

class WoodCells
{
public:
	WoodCells(const Parameters &params, const QVector<QPolygonF> &activePolys = QVector<QPolygonF> (), const QVector<
		QPolygonF> &inactivePolys = QVector<QPolygonF> ());
	virtual ~WoodCells();

	void growth();

	friend class WoodSkeleton;

private:
	// Wood initialization methods
	void populatePoint(const QPointF &point);
	void populatePoly(const QPolygonF &poly);
	void populateInactive(const QPolygonF &poly);

	// General methods for cell growth
	void checkConfiguration();
	void recordConfiguration();

	// Specific heuristics for group growth
	void initNewGroup(const CellGroup &group);
	void splitIntersections(CellGroup group, int i);
	void mergeGroup(CellGroup &group);
	void eraseIntersections(CellGroup &group) const;
	void checkCollisions(CellGroup &group1, CellGroup &group2);
	void readjustGroup(CellGroup &oldGroup);

	// Miscellaneous
	int bridgeBetween(Cell *left, Cell *right) const;
	QColor colorFromAge(Cell *cell) const;

	Q_DISABLE_COPY(WoodCells)

private:
	typedef QPair<Cell *, Cell *> Intersection;

	// Internal parameters
	double speed[4];
	double zeta;
	double distanceLimit;
	double distanceSquared;
	double angularStdev;
	double lengthStdev;
	int growthIter;
	int groupCounter;
	int stepPerCycle;

	QColor colorEarlyWood;
	QColor colorLateWood;

	QVector<CellGroup> active;
	QVector<CellGroup> inactive;

	QVector<CellGroup> initial;
	QVector<CellGroup> final;

	// Element to delete upon object destruction
	QVector<Cell *> elements;
	QVector<Cell *> invisible;
};

/*****************************************************************************
 Return the color corresponding to a given age
 *****************************************************************************/

inline QColor WoodCells::colorFromAge(Cell *cell) const {
	double t;
	if (cell->clockWise)
		t = pow(1.0 - (double) mod(cell->radius, stepPerCycle) / (double) qMax(1, stepPerCycle - 1), 1.0 / zeta);
	else
		t = pow((double) mod(cell->radius, stepPerCycle) / (double) qMax(1, stepPerCycle - 1), 1.0 / zeta);
	return QColor::fromRgbF(t * colorEarlyWood.redF() + (1 - t) * colorLateWood.redF(), t * colorEarlyWood.greenF()
		+ (1 - t) * colorLateWood.greenF(), t * colorEarlyWood.blueF() + (1 - t) * colorLateWood.blueF());
}

#endif /* WOODCELLS_H_ */
