/*
 * WoodSection.cpp
 *
 */

#include "WoodSection.h"
#include "WoodCells.h"
#include "WoodSkeleton.h"
#include "Parameters.h"
#include "Geometry.h"
#include "Random.h"
#include "Color.h"
#include <cstdio>
#include <limits>
#include <QtOpenGL>
#include <QGLViewer/frame.h>

using namespace qglviewer;

WoodSection::WoodSection(Frame *frame) :
	frame(frame), drawFlags(Active | Border | Vertices), idSelection(-1), activeLastIndex(0), inactiveLastIndex(0),
		cells(0), skeleton(0) {
}

WoodSection::~WoodSection() {
	delete cells;
	delete skeleton;
}

/*****************************************************************************
 Drawing functions
 *****************************************************************************/

void WoodSection::drawWithNames(float radius) const {
	double s = radius;

	if (drawFlags & Border) {
		// Draw active points
		qglColor(Qt::darkBlue);
		glNormal3f(0, 0, -1.f);
		for (int i = 0; i < activePoints.size(); ++i) {
			if (i == idSelection)
				qglColor(Qt::yellow);
			glPushName(i);
			glBegin(GL_QUADS);
			glVertex3d(activePoints[i].x() - s, activePoints[i].y() - s, 0.01);
			glVertex3d(activePoints[i].x() - s, activePoints[i].y() + s, 0.01);
			glVertex3d(activePoints[i].x() + s, activePoints[i].y() + s, 0.01);
			glVertex3d(activePoints[i].x() + s, activePoints[i].y() - s, 0.01);
			glEnd();
			glPopName();
			if (i == idSelection)
				qglColor(Qt::darkBlue);
		}

		// Draw inactive points
		qglColor(Qt::darkRed);
		int offset = activePoints.size();
		for (int i = 0; i < inactivePoints.size(); ++i) {
			if (i + offset == idSelection)
				qglColor(Qt::yellow);
			glPushName(i + offset);
			glBegin(GL_QUADS);
			glVertex3d(inactivePoints[i].x() - s, inactivePoints[i].y() - s, 0.01);
			glVertex3d(inactivePoints[i].x() - s, inactivePoints[i].y() + s, 0.01);
			glVertex3d(inactivePoints[i].x() + s, inactivePoints[i].y() + s, 0.01);
			glVertex3d(inactivePoints[i].x() + s, inactivePoints[i].y() - s, 0.01);
			glEnd();
			glPopName();
			if (i + offset == idSelection)
				qglColor(Qt::darkRed);
		}
	}
}

void WoodSection::draw(float radius) const {
	glNormal3f(0, 0, -1.f);
	drawGroup(activePoints, activeIndex, Qt::darkBlue, idSelection, radius);
	drawGroup(inactivePoints, inactiveIndex, Qt::darkRed, idSelection - activePoints.size(), radius);
	if (skeleton)
		skeleton->draw(drawFlags & Wireframe ? true : false);

	return;

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	qglColor(Qt::darkMagenta);
	QVector<Triplet> triangles;
	QVector<QVector3D> points (activePoints.size());
	for (int i = 0; i < activePoints.size(); ++i)
		points[i] = QVector3D(activePoints[i].x(), activePoints[i].y(), 0.01f);

	triangulatePolygon(activePoints, triangles);

	// Enable and load arrays
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(2, GL_DOUBLE, 0, activePoints.constData());

	// Draw triangles
	glDrawElements(GL_TRIANGLES, 3 * triangles.size(), GL_UNSIGNED_INT, triangles.constData());

	// Restore states
	glDisableClientState(GL_VERTEX_ARRAY);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void WoodSection::drawGroup(const QVector<QPointF> &points, const QVector<int> &index, Qt::GlobalColor vertexColor,
	int localSelection, float s) const {

	// Draw current points
	if (drawFlags & Vertices) {
		qglColor(vertexColor);
		glBegin(GL_QUADS);
		for (int i = 0; i < points.size(); ++i) {
			if (i == localSelection)
				qglColor(Qt::yellow);
			glVertex3d(points[i].x() - s, points[i].y() - s, 0.01);
			glVertex3d(points[i].x() - s, points[i].y() + s, 0.01);
			glVertex3d(points[i].x() + s, points[i].y() + s, 0.01);
			glVertex3d(points[i].x() + s, points[i].y() - s, 0.01);
			if (i == localSelection)
				qglColor(vertexColor);
		}
		glEnd();
	}

	// Draw points border
	if (drawFlags & Border) {
		if (index.size() < 3)
			return;
		qglColor(vertexColor);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		int lastIndex = -1;
		for (int i = 0; i < points.size(); ++i) {
			if (index[i] != lastIndex) {
				lastIndex = index[i];
				if (lastIndex != -1)
					glEnd();
				glBegin(GL_POLYGON);
			}
			glVertex3d(points[i].x(), points[i].y(), 0.01);
		}
		glEnd();
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
}

/*****************************************************************************
 Miscellaneous
 *****************************************************************************/

float WoodSection::computeRadius() const {
	double radius = 0;
	for (int i = 0; i < activePoints.size(); ++i) {
		Vec point(activePoints[i].x(), activePoints[i].y(), 0);
		radius = qMax(radius, (double) point.squaredNorm());
	}
	for (int i = 0; i < inactivePoints.size(); ++i) {
		Vec point(inactivePoints[i].x(), inactivePoints[i].y(), 0);
		radius = qMax(radius, (double) point.squaredNorm());
	}
	radius = sqrt(radius);
	if (skeleton)
		return qMax((float) radius, skeleton->computeRadius());
	else
		return radius;
}

void WoodSection::growth(const Parameters *params) {
	// Remove old elements
	delete cells;
	delete skeleton;

	// Initialize random number generator
	rng.setSeed(params->randomSeed);

	// Initialize the wood cells
	QVector<QPolygonF> activePolys;
	QVector<QPolygonF> inactivePolys;
	buildPolygons(activePoints, activeIndex, activePolys);
	buildPolygons(inactivePoints, inactiveIndex, inactivePolys);
	cells = new WoodCells(*params, activePolys, inactivePolys);

	// Primary growth
	cells->growth();

	// Build the resulting skeleton
	skeleton = new WoodSkeleton(cells);
	objectChanged();
}

void WoodSection::selectNearest(const QPointF &pos) {
	double minLengthSquared = std::numeric_limits<double>::max();
	idSelection = -1;

	for (int i = 0; i < activePoints.size(); ++i) {
		double tmpLengthSquared = lengthSquared(activePoints[i] - pos);
		if (tmpLengthSquared < minLengthSquared) {
			minLengthSquared = tmpLengthSquared;
			idSelection = i;
		}
	}

	for (int i = 0; i < inactivePoints.size(); ++i) {
		double tmpLengthSquared = lengthSquared(inactivePoints[i] - pos);
		if (tmpLengthSquared < minLengthSquared) {
			minLengthSquared = tmpLengthSquared;
			idSelection = i + activePoints.size();
		}
	}

	objectChanged();
}

void WoodSection::clear() {
	delete cells;
	delete skeleton;
	cells = 0;
	skeleton = 0;
	activeLastIndex = 0;
	activeIndex.clear();
	activePoints.clear();
	inactiveLastIndex = 0;
	inactiveIndex.clear();
	inactivePoints.clear();
	objectChanged();
}

void WoodSection::removeSelection() {
	if (idSelection < 0)
		return;
	if (idSelection < activePoints.size()) {
		activePoints.remove(idSelection);
		activeIndex.remove(idSelection);
	} else {
		idSelection -= activePoints.size();
		inactivePoints.remove(idSelection);
		inactiveIndex.remove(idSelection);
	}
	idSelection = -1;
	objectChanged();
}

/*****************************************************************************
 Loading and saving
 *****************************************************************************/

void WoodSection::loadOneGroup(QTextStream &in, QVector<QPointF> &poly, QVector<int> &index, int &lastIndex) const {
	int numPoly = 0;
	int sizePoly = 0;

	for (in >> numPoly; numPoly; --numPoly) {
		// Read one polygon
		for (in >> sizePoly; sizePoly; --sizePoly) {
			QPointF pts;
			in >> pts.rx() >> pts.ry();
			poly.append(pts);
			index.append(lastIndex);
		}
		++lastIndex;
	}
}

void WoodSection::load(const QString &fileName) {
	clear();

	// Open the file
	QFile file(fileName);
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		return;
	QTextStream in(&file);

	// Load the polygons
	loadOneGroup(in, activePoints, activeIndex, activeLastIndex);
	loadOneGroup(in, inactivePoints, inactiveIndex, inactiveLastIndex);
	objectChanged();

	printf("Polygons loaded successfully\n");
}

void WoodSection::saveOneGroup(QTextStream &out, const QVector<QPointF> &poly, const QVector<int> &index) const {
	// Special case: empty group
	if (poly.isEmpty()) {
		out << 0 << endl;
		return;
	}

	// Retrieve the polygon sizes
	QVector<int> sizePoly;
	int current = 0;
	for (int i = 0, acc = index.first(); i < poly.size(); ++i, ++current) {
		if (index[i] != acc) {
			acc = index[i];
			sizePoly.append(current);
			current = 0;
		}
	}
	sizePoly.append(current);

	// Write to the output text file
	out << sizePoly.size() << endl;
	for (int i = 0, j = 0, k = 0; i < sizePoly.size(); ++i) {
		// Write one polygon
		out << sizePoly[i] << endl;
		for (j = 0; j < sizePoly[i]; ++j, ++k)
			out << poly[k].x() << " " << poly[k].y() << endl;
	}
}

void WoodSection::save(const QString &fileName) const {
	// Open the file
	QFile file(fileName);
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
		return;
	QTextStream out(&file);

	// Save the polygons
	saveOneGroup(out, activePoints, activeIndex);
	saveOneGroup(out, inactivePoints, inactiveIndex);

	printf("Polygons saved successfully\n");
}

/*****************************************************************************
 Internal methods
 *****************************************************************************/

void WoodSection::buildPolygons(const QVector<QPointF> &poly, const QVector<int> index, QVector<QPolygonF> &ans) const {
	if (index.isEmpty())
		return;

	ans.resize(ans.size() + 1);
	for (int i = 0, acc = index.first(); i < poly.size(); ++i) {
		if (index[i] != acc) {
			acc = index[i];
			ans.resize(ans.size() + 1);
		}
		ans.last().append(poly[i]);
	}
}
