/*
 * WoodSection.h
 *
 */

#ifndef WOODSECTION_H_
#define WOODSECTION_H_

// Include
#include "Object.h"
#include <QString>
#include <QColor>
#include <QPointF>
#include <QVector>
#include <QPolygonF>
#include <QTextStream>

// Class declaration
namespace qglviewer {
	class Frame;
}
class WoodCells;
class WoodSkeleton;
class Parameters;

// Class definition
class WoodSection: public Object
{
	typedef qglviewer::Frame Frame;

public:
	enum DrawFlags
	{
		Default = 0, Active = 1, Former = 2, Border = 4, Vertices = 8, Skeleton = 16, Wireframe = 32,
	};

public:
	WoodSection(Frame *frame);
	virtual ~WoodSection();

	void drawWithNames(float radius = 1.f) const;
	void draw(float radius = 1.f) const;

	void growth(const Parameters *params);
	void clear();
	void load(const QString &fileName);
	void save(const QString &fileName) const;

	void select(int id);
	void selectNearest(const QPointF &pos);
	void moveSelection(const QPointF &newPos);
	void removeSelection();

	void createGroup(bool active = true);
	void createVertex(const QPointF &pos, bool active = true);
	void toggleInfo(int flag);

	float computeRadius() const;
	const Frame* plane() const;

protected:
	void drawGroup(const QVector<QPointF> &poly, const QVector<int> &index, Qt::GlobalColor vertexColor,
		int localSelection, float s) const;

	void loadOneGroup(QTextStream &in, QVector<QPointF> &poly, QVector<int> &index, int &lastIndex) const;
	void saveOneGroup(QTextStream &out, const QVector<QPointF> &poly, const QVector<int> &index) const;

	void buildPolygons(const QVector<QPointF> &poly, const QVector<int> index, QVector<QPolygonF> &ans) const;

private:
	Frame* frame;
	int drawFlags;
	int idSelection;

	int activeLastIndex;
	int inactiveLastIndex;

	QVector<int> activeIndex;
	QVector<int> inactiveIndex;

	QVector<QPointF> activePoints;
	QVector<QPointF> inactivePoints;

	WoodCells *cells;
	WoodSkeleton *skeleton;
};

// Inline methods
inline const qglviewer::Frame * WoodSection::plane() const {
	return frame;
}

inline void WoodSection::select(int id) {
	idSelection = id;
	objectChanged();
}

inline void WoodSection::moveSelection(const QPointF &newPos) {
	if (idSelection < 0)
		return;
	if (idSelection < activePoints.size())
		activePoints[idSelection] = newPos;
	else
		inactivePoints[idSelection - activePoints.size()] = newPos;
	objectChanged();
}

inline void WoodSection::createGroup(bool active) {
	if (active) {
		if (not activeIndex.isEmpty() && activeIndex.last() == activeLastIndex)
			++activeLastIndex;
	} else {
		if (not inactiveIndex.isEmpty() && inactiveIndex.last() == inactiveLastIndex)
			++inactiveLastIndex;
	}
}

inline void WoodSection::createVertex(const QPointF &pos, bool active) {
	if (active) {
		activePoints.append(pos);
		activeIndex.append(activeLastIndex);
	} else {
		inactivePoints.append(pos);
		inactiveIndex.append(inactiveLastIndex);
	}
	idSelection = -1;
	objectChanged();
}

inline void WoodSection::toggleInfo(int flag) {
	drawFlags ^= flag;
	objectChanged();
}

#endif /* WOODSECTION_H_ */
