/*
 * WoodSkeleton.cpp
 *
 */

#include "WoodSkeleton.h"
#include "WoodCells.h"
#include "CellGroup.h"
#include "Geometry.h"
#include <cstdio>
#include <QtOpenGL>
#include <QGLViewer/vec.h>

WoodSkeleton::WoodSkeleton(const WoodCells *cells) {
	if (cells) {
		buildFaces(*cells);
		buildTriangles(*cells);
	}
}

WoodSkeleton::~WoodSkeleton() {

}

/*****************************************************************************
 Building the skeleton
 *****************************************************************************/

void WoodSkeleton::buildFaces(const WoodCells &cells) {
	// Create vertices
	vertices.resize(cells.elements.size());
	colors.resize(cells.elements.size());
	for (int i = 0; i < cells.elements.size(); ++i) {
		const QPointF &point = cells.elements[i]->position;
		const QColor color = cells.colorFromAge(cells.elements[i]);
		vertices[i] = QVector3D(point.x(), point.y(), 0.02f);
		if (cells.elements[i]->isDead())
			colors[i] = QVector3D();
		else
			colors[i] = QVector3D(color.redF(), color.greenF(), color.blueF());
		cells.elements[i]->index = i;
	}

	// Create polygonal faces
	for (int i = 0; i < cells.elements.size(); ++i) {
		Cell *cell = cells.elements[i];
		if (cell->child) {
			Cell *left = cell->left;
			while (not left->child)
				left = left->left;
			if (cell->child->group == left->child->group) {
				//&& not (cell->child->state == Cell::Dead || left->child->state == Cell::Dead)) {
				faces.append(QVector<uint> ());
				Cell *current = cell;
				while (current != left) {
					faces.last().append(current->index);
					current = current->left;
				}
				faces.last().append(current->index);
				current = current->child;
				while (current != cell->child) {
					faces.last().append(current->index);
					current = current->right;
				}
				faces.last().append(current->index);
			}
		}
	}

	// Special loop for initial groups oriented outward
	buildPolygons(cells.initial, true);

	// Special loop for final groups oriented inward
	buildPolygons(cells.active, false);
	buildPolygons(cells.inactive, false);
	buildPolygons(cells.final, false);
}

/*****************************************************************************
 Building a polygon from a group
 *****************************************************************************/

void WoodSkeleton::buildPolygons(const QVector<CellGroup> &groups, bool orientation) {
	for (int i = 0; i < groups.size(); ++i)
		if (groups[i].orientation == orientation && groups[i].start->index != -1) {
			faces.append(QVector<uint> ());
			for (Cell *cell = groups[i].start->right;; cell = cell->right) {
				faces.last().append(cell->index);
				if (cell == groups[i].start)
					break;
			}
		}
}

/*****************************************************************************
 Building the triangles
 *****************************************************************************/

void WoodSkeleton::buildTriangles(const WoodCells &cells) {
	for (int i = 0; i < faces.size(); ++i) {
		QPolygonF poly(faces[i].size());
		for (int j = 0; j < faces[i].size(); ++j)
			poly[j] = cells.elements[faces[i][j]]->position;
		QVector<Triplet> tmp;
		triangulatePolygon(poly, tmp);
		for (int k = 0; k < tmp.size(); ++k)
			triangles.append(Triplet(faces[i][tmp[k][0]], faces[i][tmp[k][1]], faces[i][tmp[k][2]]));
	}
}

/*****************************************************************************
 Radius estimation
 *****************************************************************************/

float WoodSkeleton::computeRadius() const {
	qreal radius = 0;
	for (int i = 0; i < vertices.size(); ++i)
		radius = qMax(radius, vertices[i].length());
	return radius;
}

/*****************************************************************************
 Drawing function
 *****************************************************************************/

void WoodSkeleton::draw(bool wireframe) const {
	// Enable and load arrays
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, vertices.constData());

	glEnableClientState(GL_COLOR_ARRAY);
	glColorPointer(3, GL_FLOAT, 0, colors.constData());

	if (not wireframe) {
		// Draw triangles
		glDrawElements(GL_TRIANGLES, 3 * triangles.size(), GL_UNSIGNED_INT, triangles.constData());
	} else {
		// Draw faces
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		for (int i = 0; i < faces.size(); ++i)
			glDrawElements(GL_POLYGON, faces[i].size(), GL_UNSIGNED_INT, faces[i].constData());
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}

	// Restore states
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
}
